/*
Copyright (C) 1998-2007 Valerio Olevano, Lucia Reining, Francesco Sottile, CNRS.
All rights reserved.

DP code
Linear Response TDDFT code in Frequency-Reciprocal space on a Plane Waves basis set.
Calculate Dielectric Properties.
*/

/* 
fermi library 
calculate fermi energy and fermi level occupation
for the spin polarized and unpolarized cases       
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "ggw.h"
#define EPS_DEG 0.00001 /* difference in the levels to consider degenerate [Hartree] */

int is_insulator(int *, int *, int *, int *, float *, double *, double *, double *, double *);
double fermi_energy(int *, int *, int *, int *, float *, float *);
double fermi_occupation(int *, int *, int *, int *, float *, float *);


struct level { double e; double w;};
int compare_levels(const void *, const void *);


is_insulator__(int *ns, int *nb, int *nk, int *nel, float *e, 
             double *tve, double *bce, double *mge, double *dge)
{ return(is_insulator(ns,nb,nk,nel,e,tve,bce,mge,dge)) ;}
is_insulator_(int *ns, int *nb, int *nk, int *nel, float *e, 
             double *tve, double *bce, double *mge, double *dge)
{ return(is_insulator(ns,nb,nk,nel,e,tve,bce,mge,dge)) ;}
is_insulator(int *ns, int *nb, int *nk, int *nel, float *e, 
             double *tve, double *bce, double *mge, double *dge)
/* verify if there is a gap, i.e. if it is an insulator
   nota: the bands must be ordered for increasing energy
   though the spin degree of freedom does not need
   ns: number of spins (2 spin polarized, 1 unpolarized)
   nb: number of bands
   nk: number of k-points
   nel: number of electrons
   f90: e(nk,nb,ns) ; C: e[ns][nb][nk]: energies
   return values:
   1 (true) if insulator
   0 if semimetal
   -1 if metal 
   tve: top of valence energy
   bce: bottom of conduction energy
   mge: minimum gap energy
   dge: direct gap energy
*/
{
  int lumo, homo;
  int k, i_lumodown, i_lumoup, i_homodown, i_homoup;
  int ins;
  double *ehomo, *elumo;

  lumo = *nel / 2;
  homo = (*nel-1) / 2;
  if(homo == lumo) return(ins=-1); /* if nel is odd then it is a metal */

  /* initialize tve, bce, dge */
  *tve = e[0];
  *bce = e[*nk * *nb * *ns - 1];
  *dge = *bce - *tve;

  ehomo = (double*) calloc(*nk,sizeof(double));
  elumo = (double*) calloc(*nk,sizeof(double));
  for(k=0;k<*nk;k++) {
    i_lumodown = k + lumo * *nk;
    i_lumoup   = k + lumo * *nk + (*ns-1) * *nb * *nk;
    i_homodown = k + homo * *nk;
    i_homoup   = k + homo * *nk + (*ns-1) * *nb * *nk;
    ehomo[k] = (e[i_homoup] > e[i_homodown]) ? e[i_homoup] : e[i_homodown];
    elumo[k] = (e[i_lumodown] < e[i_lumoup]) ? e[i_lumodown] : e[i_lumoup];
    *dge = ((elumo[k]-ehomo[k]) < *dge) ? elumo[k]-ehomo[k] : *dge;
    *tve = (ehomo[k] > *tve) ? ehomo[k] : *tve;
    *bce = (elumo[k] < *bce) ? elumo[k] : *bce;
  }

  ins = *bce > *tve; /* = true = 1 if insulator */
  if(*bce < *tve) ins = -1; /* metal */
  if(*dge == 0.0) ins = -2; /* this can be a metal or a semimetal;
                               indeed there is at least a point where the 
                               last occ band and the first unocc band touch;
                               if tve == bce at the point dge=0
                               then it is a semimetal;
                               if tve > bce then 
                               if the fermi energy fall exactly
                               in the point (or in a line) where dge=0 
                               then it is a semimetal;
                               otherwise the fermi energy fall in a point
                               where there is a direct gap, and it is a metal.*/
  if(*bce == *tve) ins = 0; /* semimetal */
  *mge = *bce - *tve;
  free(ehomo);
  free(elumo);
  return(ins);  
}


double fermi_energy__(int *ns, int *nb, int *nk, int *nel, float *e, float *w)
{ return(fermi_energy(ns,nb,nk,nel,e,w)); }
double fermi_energy_(int *ns, int *nb, int *nk, int *nel, float *e, float *w)
{ return(fermi_energy(ns,nb,nk,nel,e,w)); }
double fermi_energy(int *ns, int *nb, int *nk, int *nel, float *e, float *w)
/* calculate fermi energy
   ns: number of spins (2 spin polarized, 1 unpolarized)
   nb: number of bands
   nk: number of k-points
   nel: number of electrons
   f90: e(nk,nb,ns) ; C: e[ns][nb][nk]: energies
   w[nk]: k-points weights   */
{
  struct level *l;
  int n;
  double fe, fo;
  double tve, bce, mge, dge;

  if(is_insulator(ns,nb,nk,nel,e,&tve,&bce,&mge,&dge)==1) return(tve);

  n = *ns * *nb * *nk;
  l = (struct level *) calloc(n,sizeof(struct level));
  set_levels(l,*ns,*nb,*nk,*nel,e,w);
  fermi(&fe,&fo,*nel,n,l);
  free(l);
  return(fe);
}


double fermi_occupation__(int *ns, int *nb, int *nk, int *nel, float *e, float *w)
{ return(fermi_occupation(ns,nb,nk,nel,e,w)); }
double fermi_occupation_(int *ns, int *nb, int *nk, int *nel, float *e, float *w)
{ return(fermi_occupation(ns,nb,nk,nel,e,w)); }
double fermi_occupation(int *ns, int *nb, int *nk, int *nel, float *e, float *w)
/* calculate fermi level occupation:
   ns: number of spins (2 spin polarized, 1 unpolarized)
   nb: number of bands
   nk: number of k-points
   nel: number of electrons
   f90: e(nk,nb,ns) ; C: e[ns][nb][nk]: energies
   w[nk]: k-points weights
   return values:
   spin unpolarized case (ns=1): from 0 to 2
   spin polarized case   (ns=2): from 0 to 1    */
{
  struct level *l;
  int n;
  double fe, fo;
  double tve, bce, mge, dge;
  double spin_fac;

  if(*ns==1) spin_fac = 2;
  else if(*ns==2) spin_fac = 1;
  else exit(1);


  if(is_insulator(ns,nb,nk,nel,e,&tve,&bce,&mge,&dge)==1) return(spin_fac);

  n = *ns * *nb * *nk;
  l = (struct level *) calloc(n,sizeof(struct level));
  set_levels(l,*ns,*nb,*nk,*nel,e,w);
  fermi(&fe,&fo,*nel,n,l);
  free(l);

  fo *= spin_fac;
  return(fo);
}

 
fermi(double *fe, double *fo, int nel, int n, struct level *l)
/* given the unordered set of level, calculate fermi energy and occupation */
{
  int i;
  double acc=0.0;

  /* sort levels in ascending order (stdlib)*/
  qsort(l,n,sizeof(struct level),compare_levels);
  /* reduce degeneracies */
  reduce_degenerate_levels(l,&n);

  /* find last occupied level */
  for(i=0;((acc+l[i].w) < nel) && (i<n);i++) acc+=l[i].w;
  if(i>=n) return(-1);
  *fe = l[i].e;
  *fo = (nel - acc) / l[i].w;

  return(0);
}


set_levels(struct level *l, int ns, int nb, int nk, int nel, float *e, float *w)
/* setup levels in a table l */
{
  double wtot = 0;
  double spin_fac;
  int i, s, b, k;

  if(ns==1) spin_fac = 2;
  else if(ns==2) spin_fac = 1;
  else exit(1);

  for(k=0,wtot=0;k<nk;k++) wtot+=w[k];

  i = 0;
  for(s=0;s<ns;s++)
    for(b=0;b<nb;b++)
      for(k=0;k<nk;k++) {
        l[i].e = e[i];
        l[i].w = spin_fac * w[k] / wtot;
        i++;
      };
  return(0);
}


reduce_degenerate_levels(struct level *l,int *n)
/* reduce the degeneracies in the set of levels */
{
  int i, j;
  
  for(i=0;i<(*n)-1;i++) {
    if(levels_are_degenerate(&l[i],&l[i+1])) {
      l[i].w += l[i+1].w;
      for(j=i+1;j<*n-1;j++) {l[j].e = l[j+1].e; l[j].w = l[j+1].w;}
      (*n)--;
      i--;
    }
  }
  return(0);
}


int compare_levels(const void * a, const void * b)
/* compare two levels and return -1, 0, +1 if 1st energy is <, ==, > 2nd */
{
  int ret;
  ret = (((struct level *)a)->e <= ((struct level *)b)->e ) ? -1 : +1 ;
  if    (((struct level *)a)->e == ((struct level *)b)->e ) ret = 0;
  return(ret);
}


int levels_are_degenerate(struct level *a, struct level *b)
/* compare two levels and return true if they are degenerate */
{
  /* return(fabs(a->e - b->e)<EPS_DEG); */
  return(fabs(a->e - b->e)==0.0);
}


print_data_(int *is_insulator, 
           double *tve,  double *bce, double *mge, double *dge,
           double *efermi, double *ofermi)
{return(print_data(is_insulator,tve,bce,mge,dge,efermi,ofermi));}
print_data(int *is_insulator, 
           double *tve,  double *bce, double *mge, double *dge,
           double *efermi, double *ofermi)
{
  switch(*is_insulator) {
  case 1:
    printf("insulator\n");
    printf(" top of valence       %6.2f [eV]\n",*tve * H2eV);
    printf(" bottom of conduction %6.2f [eV]\n",*bce * H2eV);
    printf(" minimum band gap     %6.2f [eV]\n",*mge * H2eV);
    printf(" direct band gap      %6.2f [eV]\n",*dge * H2eV);
    break;
  case 0:
    printf("semimetal\n");
    printf(" Fermi energy         %6.2f [eV]\n",*efermi * H2eV);
    printf(" Fermi occupation     %8.4f [eV]\n",*ofermi);
    break;
  case -1:
    printf("metal\n");
    printf(" Fermi energy         %6.2f [eV]\n",*efermi * H2eV);
    printf(" Fermi occupation     %8.4f [eV]\n",*ofermi);
    break;
  }
  return(0);
}
