! Copyright (C) 1998-2007 Valerio Olevano, Lucia Reining, Francesco Sottile, CNRS.
! All rights reserved.
!
! DP code
! Linear Response TDDFT code in Frequency-Reciprocal space on a Plane Waves basis set.
! Calculate Dielectric Properties.

!     ==================================================================
!     interface to the ASL library (NEC SX5)
!     ==================================================================


!     ==================================================================
      subroutine chevc(n,a,lda,w,z,ldz)
!     ==================================================================
!     compute the eigenvalues and the eigenvectors
!     of a complex Hermitian matrix.
!     only the upper triangle of a is used
!     ------------------------------------------------------------------
      implicit none
      integer n, lda, ldz
      complex a(lda,n)
      complex z(ldz,n)
      real w(n)
      real w1(n)
      complex w2(n)
      integer ierr
      call ccheaa(a,lda,n,w,w1,w2,ierr)
      z(1:n,1:n) = a(1:n,1:n)
      return
      end


!     ==================================================================
      subroutine cgevc(n,a,lda,w,z,ldz)
!     ==================================================================
!     compute the eigenvalues and the (right) eigenvectors
!     of a complex general matrix
!     a and z must have no common elements
!     ------------------------------------------------------------------
      implicit none
      integer n, lda, ldz
      complex a(lda,n)
      complex z(ldz,n)
      complex w(n)
      real w1(n)
      complex w2(n)
      integer ierr
      call ccgnaa(a,lda,n,w,z,ldz,w1,w2,ierr)
      return
      end


!     ==================================================================
      subroutine cginv(n,a,lda)
!     ==================================================================
!     invert a complex general matrix
!     CBGNLU computes an LU factorization of a general N-by-N matrix A
!     using partial pivoting with row interchanges.
!     CBGNDI computes the inverse of a matrix using the LU factorization
!     computed by CBGNLU.
!     ------------------------------------------------------------------
      implicit none
      integer n, lda
      complex a(lda,n)
      integer ipvt(n)
      integer ierr
      complex cdet
      real det
      complex w1(n)
      integer, parameter :: isw = -1
      call cbgnlu(a,lda,n,ipvt,ierr)
      call cbgndi(a,lda,n,ipvt,cdet,det,isw,w1,ierr)
      return
      end


!     ==================================================================
      subroutine zginv(n,a,lda)
!     ==================================================================
!     invert a complex general matrix
!     ZBGNLU computes an LU factorization of a general N-by-N matrix A
!     using partial pivoting with row interchanges.
!     ZBGNDI computes the inverse of a matrix using the LU factorization
!     computed by ZBGNLU.
!     ------------------------------------------------------------------
      implicit none
      integer n, lda
      complex  (kind(0.0d0)) :: a(lda,n)
      integer ipvt(n)
      integer ierr
      complex  (kind(0.0d0)) :: cdet
      real  (kind(0.0d0)) :: det
      complex  (kind(0.0d0)) :: w1(n)
      integer, parameter :: isw = -1
      call zbgnlu(a,lda,n,ipvt,ierr)
      call zbgndi(a,lda,n,ipvt,cdet,det,isw,w1,ierr)
      return
      end


!     ==================================================================
      subroutine rginv(n,a,lda)
!     ==================================================================
!     invert a real general matrix
!     RBGMLU computes an LU factorization of a general n-by-N matrix A
!     using partial pivoting with row interchanges.
!     RBGMDI computes the inverse of a matrix using the LU factorization
!     computed by RBGMLU.
!     ------------------------------------------------------------------
      implicit none
      integer n, lda
      real a(lda,n)
      integer ipvt(n)
      real w1(n), det(2)
      integer ierr
      integer, parameter :: isw = -1
      call rbgmlu(a,lda,n,ipvt,ierr)
      call rbgmdi(a,lda,n,ipvt,det,isw,w1,ierr)
      return
      end


!     ==================================================================
      subroutine cgle(n,a,lda,b,x)
!     ==================================================================
!     solve the general complex problem A * X = B
!     where A is a complex matrix N x N and X, B are complex vectors
!     ------------------------------------------------------------------
      implicit none
      integer n, lda
      complex a(lda,n), x(n), b(n)
      
      integer ierr
      integer ipvt(n)
      complex bb(n)
      
      bb(:) = b(:)
      call cbgnsl(a,lda,n,bb,ipvt,ierr)
      if(ierr>1000) then
        print *, ierr
        stop 'cbgnsl failed'
      endif
      x(:) = bb(:)
      
      return
      end


!     ==================================================================
      subroutine zgle(n,a,lda,b,x)
!     ==================================================================
!     solve the general complex problem A * X = B
!     where A is a complex matrix N x N and X, B are complex vectors
!     ------------------------------------------------------------------
      implicit none
      integer n, lda
      complex (kind(0.0d0)) :: a(lda,n), x(n), b(n)
      
      integer ierr
      integer ipvt(n)
      complex (kind(0.0d0)) :: bb(n)
      
      bb(:) = b(:)
      call zbgnsl(a,lda,n,bb,ipvt,ierr)
      if(ierr>1000) then
        print *, ierr
        stop 'zbgnsl failed'
      endif
      x(:) = bb(:)
      
      return
      end


!     ==================================================================
!     Fourier transforms
!     ==================================================================


!     ==================================================================
      integer function allowed_fftsize(n)
!     ==================================================================
!     calculate the size of the smallest allowed FFT size that
!     will encompass n
!     this routine is written for the ESSL routine scft, in which the
!     allowable sizes of FFT are those whose prime factors are
!       n <= 37748736
!       n = 2^h * 3^i * 5^j * 7^k * 11^m  h=1,...,25  i=0,1,2  j,k,m=0,1
!     ------------------------------------------------------------------
      implicit none
      integer n
      integer, parameter :: nn = 56
      integer nallwd(nn)
      data nallwd/
     &   2,   4,   6,   8,  10,  12,  14,  16,  18,  20,
     &  22,  24,  28,  30,  32,  36,  40,  42,  44,  48,
     &  56,  60,  64,  66,  70,  72,  80,  84,  88,  90,
     &  96, 110, 112, 120, 126, 128, 132, 140, 144, 154,
     & 160, 168, 176, 180, 192, 198, 210, 220, 224, 240,
     & 252, 256, 264, 280, 288, 308/
      integer i

      if(n > nallwd(nn) .or. n < 2) then
        print *, '**ERROR allowed_fftsize'
        stop 'error allowed fft_size'
      endif
      do i = nn, 1, -1
        if(nallwd(i) >= n) allowed_fftsize = nallwd(i)
      enddo
      return
      end


!     ==================================================================
      subroutine nfft(m,n)
!     ==================================================================
!     calculate N, the size of the smallest allowed FFT that will encompass
!     the 2M+1 points -M..+M.  (Ideally, N=2M+1, but this may not be
!     an allowable FFT.)
!     this routine is written for the ESSL routine scft, in which the
!     allowable sizes of FFT are those whose prime factors are
!       n <= 37748736
!       n = 2^h * 3^i * 5^j * 7^k * 11^m  h=1,...,25  i=0,1,2  j,k,m=0,1
!     ------------------------------------------------------------------
      implicit none
      integer n, m
      integer, parameter :: nn = 56
      integer nallwd(nn)
      data nallwd/
     &   2,   4,   6,   8,  10,  12,  14,  16,  18,  20,
     &  22,  24,  28,  30,  32,  36,  40,  42,  44,  48,
     &  56,  60,  64,  66,  70,  72,  80,  84,  88,  90,
     &  96, 110, 112, 120, 126, 128, 132, 140, 144, 154,
     & 160, 168, 176, 180, 192, 198, 210, 220, 224, 240,
     & 252, 256, 264, 280, 288, 308/
      integer i, nbest
      
      nbest = 2 * m + 1
      if(nbest .gt. nallwd(nn)) then
        write(*,*) '** ERROR nfft 1: '
        stop
      endif
      if(nbest .lt. 2) then
        write(*,*) '** ERROR nfft 2: '
        stop
      endif
      do i = nn, 1, -1
        if(nallwd(i) .ge. nbest) n = nallwd(i)
      enddo
      return
      end


!     ==================================================================
      subroutine n1dfft(n1,n1d)
!     ==================================================================
!     set the first dimension n1d of the FFT 3d arrays to be n1
!     ------------------------------------------------------------------
      implicit none
      integer n1, n1d
      n1d = n1
      return
      end      


!     ==================================================================
!     subroutine cfft(n,x,y,isign)
!     ==================================================================
!     compute the Fourier trasform of a complex periodic sequence
!     forward:  isign = +1
!     backward: isign = -1
!     the array x and y may be the same
!     ------------------------------------------------------------------
!     integer n
!     complex x(n)
!     complex y(n)
!     integer isign
!     
!     return
!     end


!     ==================================================================
      subroutine c3dfft(a,lda,mda,hda,n1,n2,n3,isign)
!     ==================================================================
!     compute Fourier transform of a complex periodic three-dimensional array
!     the matrices a and a may be the same
!     ------------------------------------------------------------------
      implicit none
      integer n1, n2, n3
      integer lda, mda, hda
      complex a(lda,mda,hda)
      integer isign, isw
      
      integer, save :: n1o=0, n2o=0, n3o=0
      integer, save :: ifax(60)
      real, allocatable, save :: trigs(:)
      complex, allocatable :: wk(:)
      complex, allocatable :: c(:,:,:)
      integer ierr

!      allocate(trigs(2*(n1+n2+n3)))
      allocate(wk(lda*mda*hda))
      if(n1.ne.n1o .or. n2.ne.n2o .or. n3.ne.n3o) then
!       initialize
        allocate(trigs(2*(n1+n2+n3)),stat=ierr)
        if(ierr.ne.0) stop 'c3dfft out fo memory 1'
        allocate(c(n1,n2,n3),stat=ierr)
        if(ierr.ne.0) stop 'c3dfft out fo memory 2'
        isw = 0
        call cfc3fb(n1,n2,n3,c,n1,n2,n3,isw,ifax,trigs,wk,ierr)
        if(ierr.ne.0) then
          print *, 'c3dfft error', ierr
          stop
        endif
        deallocate(c)
        n1o = n1 ; n2o = n2 ; n3o = n3
      endif
        
      
      call cfc3bf(n1,n2,n3,a,lda,mda,hda,isign,ifax,trigs,wk,ierr)

      deallocate(wk)
      
      return
      end


