! Copyright (C) 1998-2007 Valerio Olevano, Lucia Reining, Francesco Sottile, CNRS.
! All rights reserved.
!
! DP code
! Linear Response TDDFT code in Frequency-Reciprocal space on a Plane Waves basis set.
! Calculate Dielectric Properties.

!     ==================================================================
!     interface to the IMSL library (IMSLM)
!     ==================================================================


!     ==================================================================
      subroutine chevc(n,a,lda,w,z,ldz)
!     ==================================================================
!     Compute the eigenvalues and the eigenvectors
!     of a complex Hermitian matrix
!
!     only the upper triangle of a is used
!
!     Informational error
!     Type Code
!     3    1   The matrix is not Hermitian.  It has a diagonal entry
!              with a small imaginary part.
!     4    1   The iteration for an eigenvalue failed to converge.
!     4    2   The matrix is not Hermitian.  It has a diagonal entry
!              with an imaginary part.
!     ------------------------------------------------------------------
      
      integer n, lda, ldz
      complex a(lda,n)
      complex z(ldz,n)
      real w(n)
      integer i
      
      do i = 1, n
        a(i,i) = cmplx(real(a(i,i)),0.0)
      enddo
      
      call evchf(n,a,lda,w,z,ldz)
      
      return
      end


!     ==================================================================
      subroutine chevcp(n,nevec,a,lda,small,w,z,ldz)
!     ==================================================================
!     compute the largest or smallest eigenvalues and the
!     corresponding eigenvectors of a complex Hermitian matrix
!     
!     small = .true. the lowest eigenvalues are computed
!     only the upper triangle of a is used
!
!     informational errors
!     Type Code
!     3    1    The iteration for an eigenvalue failed to converge.
!               The best estimate will be returned.
!     3    2    The iteration for an eigenvector failed to converge.
!               The eigenvector will be set to 0.
!     3    3    The matrix is not Hermitian.  It has a diagonal entry
!               with a small imaginary part.
!     4    2    The matrix is not Hermitian.  It has a diagonal entry
!               with an imaginary part.
!     ------------------------------------------------------------------
      
      integer n, nevec, lda, ldz
      logical small
      complex a(lda,n)
      complex z(ldz,nevec)
      real w(n)
      integer i
      
      do i = 1, n
        a(i,i) = cmplx(real(a(i,i)),0.0)
      enddo
      
      call evehf(n,nevec,a,lda,small,w,z,ldz)
      
      return
      end


!     ==================================================================
      subroutine cgevc(n,a,lda,w,z,ldz)
!     ==================================================================
!     Compute all the eigenvalues and the (right) eigenvectors
!     of a complex general matrix
!
!     Informational error
!     Type Code
!     4    1  The iteration for the eigenvalues failed to converge.
!             No eigenvalues or eigenvectors are computed.
!     ------------------------------------------------------------------
      
      integer n, lda, ldz
      complex a(lda,n)
      complex z(ldz,n)
      complex w(n)
      
      call evccg (n,a,lda,w,z,ldz)
      return
      end


!     ==================================================================
      subroutine cginv(n,a,lda)
!     ==================================================================
!     invert a complex general matrix
!
!     Informational errors
!     Type Code
!     3    1  The input matrix is too ill-conditioned.  The inverse
!             might not be accurate.
!     4    2  The input matrix is singular.
!     ------------------------------------------------------------------
      
      integer n, lda
      complex a(lda,n)
      call lincg(n,a,lda,a,lda)
      return
      end


!     ==================================================================
      subroutine rginv(n,a,lda)
!     ==================================================================
!     invert a real general matrix
!
!     Informational errors
!     Type Code
!     3    1  The input matrix is too ill-conditioned.  The inverse
!             might not be accurate.
!     4    2  The input matrix is singular.
!     ------------------------------------------------------------------
      
      integer n, lda
      real a(lda,n)
      call linrg(n,a,lda,a,lda)
      return
      end



!     ==================================================================
!     Fourier transforms
!     ==================================================================

!     ==================================================================
      integer function allowed_fftsize(n)
!     ==================================================================
!     calculate the size of the smallest allowed FFT size that
!     will encompass n
!     this routine is written for the ESSL routine scft, in which the
!     allowable sizes of FFT are those whose prime factors are
!       n <= 37748736
!       n = 2^h * 3^i * 5^j * 7^k * 11^m  h=1,...,25  i=0,1,2  j,k,m=0,1
!     ------------------------------------------------------------------
      implicit none
      integer n
      integer, parameter :: nn = 56
      integer nallwd(nn)
      data nallwd/
     &   2,   4,   6,   8,  10,  12,  14,  16,  18,  20,
     &  22,  24,  28,  30,  32,  36,  40,  42,  44,  48,
     &  56,  60,  64,  66,  70,  72,  80,  84,  88,  90,
     &  96, 110, 112, 120, 126, 128, 132, 140, 144, 154,
     & 160, 168, 176, 180, 192, 198, 210, 220, 224, 240,
     & 252, 256, 264, 280, 288, 308/
      integer i

      if(n > nallwd(nn) .or. n < 2) then
        print *, '**ERROR allowed_fftsize'
        stop 'error allowed fft_size'
      endif
      do i = nn, 1, -1
        if(nallwd(i) >= n) allowed_fftsize = nallwd(i)
      enddo
      return
      end


!     ==================================================================
      subroutine nfft(m,n)
!     ==================================================================
!     calculate N, the size of the smallest allowed FFT that will encompass
!     the 2M+1 points -M..+M.  (Ideally, N=2M+1, but this may not be
!     an allowable FFT.)
!     this routine is written for the ESSL routine scft, in which the
!     allowable sizes of FFT are those whose prime factors are
!       n <= 37748736
!       n = 2^h * 3^i * 5^j * 7^k * 11^m  h=1,...,25  i=0,1,2  j,k,m=0,1
!     ------------------------------------------------------------------
      implicit none
      integer n, m
      integer, parameter :: nn = 56
      integer nallwd(nn)
      data nallwd/
     &   2,   4,   6,   8,  10,  12,  14,  16,  18,  20,
     &  22,  24,  28,  30,  32,  36,  40,  42,  44,  48,
     &  56,  60,  64,  66,  70,  72,  80,  84,  88,  90,
     &  96, 110, 112, 120, 126, 128, 132, 140, 144, 154,
     & 160, 168, 176, 180, 192, 198, 210, 220, 224, 240,
     & 252, 256, 264, 280, 288, 308/
      integer i, nbest
      
      nbest = 2 * m + 1
      if(nbest .gt. nallwd(nn)) then
        write(*,*) '** ERROR nfft 1: '
        stop
      endif
      if(nbest .lt. 2) then
        write(*,*) '** ERROR nfft 2: '
        stop
      endif
      do i = nn, 1, -1
        if(nallwd(i) .ge. nbest) n = nallwd(i)
      enddo
      return
      end


!     ==================================================================
      subroutine n1dfft(n1,n1d)
!     ==================================================================
!     set the first dimension n1d of the FFT 3d arrays to be n1
!     ------------------------------------------------------------------
      implicit none
      integer n1, n1d
      n1d = n1
      return
      end      


!     ==================================================================
      subroutine cfft(n,x,y,isign)
!     ==================================================================
!     compute the Fourier trasform of a complex periodic sequence
!     forward:  isign = +1
!     backward: isign = -1
!     the array x and y may be the same
!     ------------------------------------------------------------------
      integer n
      complex x(n)
      complex y(n)
      integer isign
      
      if(isign .eq. +1) then
        call fftcf(n,x,y)
      else
        call fftcb(n,x,y)
      endif
      
      return
      end


!     ==================================================================
      subroutine c3dfft(a,lda,mda,hda,n1,n2,n3,isign)
!     ==================================================================
!     compute Fourier transform of a complex periodic three-dimensional array
!     the matrices a and a may be the same
!     ------------------------------------------------------------------
      integer n1, n2, n3
      integer lda, mda, hda
      complex a(lda,mda,*)
      integer isign
      
      if(isign .eq. +1) then
        call fft3f(n1,n2,n3,a,lda,mda,a,lda,mda)
      else
        call fft3b(n1,n2,n3,a,lda,mda,a,lda,mda)
      endif
      
      return
      end


