! Copyright (C) 1998-2007 Valerio Olevano, Lucia Reining, Francesco Sottile, CNRS.
! All rights reserved.
!
! DP code
! Linear Response TDDFT code in Frequency-Reciprocal space on a Plane Waves basis set.
! Calculate Dielectric Properties.

!     ==================================================================
!     Interface to the LAPACK library
!     ==================================================================

!     ==================================================================
!     Eigensystem analysis
!     ==================================================================


!     ==================================================================
      subroutine chevc(n,a,lda,w,z,ldz)
!     ==================================================================
!     Compute the eigenvalues and the eigenvectors
!     of a complex Hermitian matrix
!     only the upper triangle of a is used
!     ------------------------------------------------------------------
      
      integer n, lda, ldz
      complex a(lda,n)
      complex z(ldz,n)
      real w(n)
      
!     upper-packed storage mode
      character, parameter :: uplo = 'U'
      character, parameter :: jobz = 'V'
      integer lwork
      real, allocatable :: rwork(:)
      complex, allocatable :: work(:)
      integer info
      
      integer istat
      
      allocate(rwork(max(1,3*n-2)),stat=istat)
      if(istat.ne.0) stop 'chevc 1: out of memory!'
      
      lwork = max(1,2*n-1)
      allocate(work(lwork),stat=istat)
      if(istat.ne.0) stop 'chevc 2: out of memory!'
!     the 6 next lines: lwork query (does not go on nec?)
!     lwork = -1
!     call cheev(jobz,uplo,n,a,lda,w,work,lwork,rwork,info)
!     lwork = work(1)
!     deallocate(work)
!     allocate(work(lwork),stat=istat)
!     if(istat.ne.0) stop 'chevc 3: out of memory!'
      
     
      call cheev(jobz,uplo,n,a,lda,w,work,lwork,rwork,info)
      do i = 1, n
        do j = 1, n
          z(i,j) = a(i,j)
        enddo
      enddo

      deallocate(rwork)
      deallocate(work)
      return
      end


!     ==================================================================
      subroutine zhevc(n,a,lda,w,z,ldz)
!     ==================================================================
!     Compute the eigenvalues and the eigenvectors
!     of a complex double precision Hermitian matrix
!     only the upper triangle of a is used
!     ------------------------------------------------------------------
      
      integer n, lda, ldz
      complex (kind(0.0d0)) :: a(lda,n)
      complex (kind(0.0d0)) :: z(ldz,n)
      double precision :: w(n)
      stop
      return
      end


!     ==================================================================
      subroutine chevcp(n,m,a,lda,small,w,z,ldz)
!     ==================================================================
!     compute the largest or smallest eigenvalues and the
!     corresponding eigenvectors of a complex Hermitian matrix
!     
!     small = .true. the lowest eigenvalues are computed
!     only the upper triangle of a is used
!     ------------------------------------------------------------------
      
      integer n, m, lda, ldz
      logical small
      complex a(lda,n)
      complex z(ldz,m)
      real w(n)
      stop
      return
      end


!     ==================================================================
      subroutine zhevcp(n,m,a,lda,small,w,z,ldz)
!     ==================================================================
!     compute the largest or smallest eigenvalues and the
!     corresponding eigenvectors of a complex double prec. Hermitian matrix
!     
!     small = .true. the lowest eigenvalues are computed
!     only the upper triangle of a is used
!     ------------------------------------------------------------------
      
      integer n, m, lda, ldz
      logical small
      complex (kind(0.0d0)) :: a(lda,n)
      complex (kind(0.0d0)) :: z(ldz,m)
      double precision :: w(n)
      stop
      return
      end


!     ==================================================================
      subroutine cgevc(n,a,lda,w,z,ldz)
!     ==================================================================
!     Compute the eigenvalues and the (right) eigenvectors
!     of a complex general matrix
!     a and z must have no common elements
!     ------------------------------------------------------------------
      
      integer n, lda, ldz
      complex a(lda,n)
      complex z(ldz,n)
      complex w(n)
      
      character, parameter :: jvl = 'N'
      character, parameter :: jvr = 'V'
      integer info
      integer lwork
      complex, allocatable :: work(:)
      real, allocatable :: rwork(:)
      integer, parameter :: ldvl = 1
      complex vl(1,1)

      integer istat
            
      lwork = max(1,2*n)
      allocate(work(lwork),stat=istat)
      if(istat.ne.0) stop 'cgevc 1: out of memory!'
      allocate(rwork(2*n),stat=istat)
      if(istat.ne.0) stop 'cgevc 2: out of memory!'

      call cgeev(jvl,jvr,n,a,lda,w,vl,ldvl,z,ldz,work,lwork,rwork,info)
      
      deallocate(work,rwork)
      return
      end


!     ==================================================================
      subroutine cginv(n,a,lda)
!     ==================================================================
!     invert a complex general matrix
!     CGETRF computes an LU factorization of a general M-by-N matrix A
!     using partial pivoting with row interchanges.
!     CGETRI computes the inverse of a matrix using the LU factorization
!     computed by CGETRF.
!     ------------------------------------------------------------------
      
      integer n, lda
      complex a(lda,n)
      
      integer ipiv(n)
      complex work(n)
      integer info
      
      call cgetrf(n,n,a,lda,ipiv,info)
      if(info.ne.0) then
        print *, info
        stop 'cgetrf failed'
      endif
      call cgetri(n,a,n,ipiv,work,n,info)
      if(info.ne.0) then
        print *, info
        stop 'cgetri failed'
      endif
      
      return
      end

!     ==================================================================
      subroutine zginv(n,a,lda)
!     ==================================================================
!     invert a complex general matrix
!     CGETRF computes an LU factorization of a general M-by-N matrix A
!     using partial pivoting with row interchanges.
!     CGETRI computes the inverse of a matrix using the LU factorization
!     computed by CGETRF.
!     ------------------------------------------------------------------
      
      integer n, lda
      complex (kind(0.0d0)) :: a(lda,n)
      
      integer ipiv(n)
      complex (kind(0.0d0)) :: work(n)
      integer info
      
      call zgetrf(n,n,a,lda,ipiv,info)
      if(info.ne.0) then
        print *, info
        stop 'cgetrf failed'
      endif
      call zgetri(n,a,n,ipiv,work,n,info)
      if(info.ne.0) then
        print *, info
        stop 'cgetri failed'
      endif
      
      return
      end


!     ==================================================================
      subroutine rginv(n,a,lda)
!     ==================================================================
!     invert a real general matrix
!     SGETRF computes an LU factorization of a general M-by-N matrix A
!     using partial pivoting with row interchanges.
!     SGETRI computes the inverse of a matrix using the LU factorization
!     computed by SGETRF.
!     ------------------------------------------------------------------
      
      integer n, lda
      real a(lda,n)
      
      integer ipiv(n)
      real work(n)
      integer info
      
      call sgetrf(n,n,a,lda,ipiv,info)
      if(info.ne.0) stop 'sgetrf failed'
      call sgetri(n,a,n,ipiv,work,n,info)
      if(info.ne.0) stop 'sgetri failed'
      
      return
      end


!     ==================================================================
      subroutine cgle(n,a,lda,b,x)
!     ==================================================================
!     solve the general complex problem A * X = B
!     where A is a complex matrix N x N and X, B are complex vectors
!     ------------------------------------------------------------------
      implicit none
      integer n, lda
      complex a(lda,n), x(n), b(n)
      
      integer, parameter :: nrhs=1
      integer info
      integer ipiv(n)
      complex bb(n,nrhs)
      
      bb(:,1) = b(:)
      call cgesv(n,nrhs,a,lda,ipiv,bb,n,info)
      if(info/=0) then
        print *, info
        stop 'cgesv failed'
      endif
      x(:) = bb(:,1)
      
      return
      end

!     ==================================================================
      subroutine zgle(n,a,lda,b,x)
!     ==================================================================
!     solve the general complex problem A * X = B
!     where A is a complex matrix N x N and X, B are complex vectors
!     ------------------------------------------------------------------
      implicit none
      integer n, lda
      complex (kind(0.0d0)) :: a(lda,n), x(n), b(n)
      
      integer, parameter :: nrhs=1
      integer info
      integer ipiv(n)
      complex (kind(0.0d0)) :: bb(n,nrhs)
      
      bb(:,1) = b(:)
      call zgesv(n,nrhs,a,lda,ipiv,bb,n,info)
      if(info/=0) then
        print *, info
        stop 'cgesv failed'
      endif
      x(:) = bb(:,1)
      
      return
      end
