! Copyright (C) 1998-2007 Valerio Olevano, Lucia Reining, Francesco Sottile, CNRS.
! All rights reserved.
!
! DP code
! Linear Response TDDFT code in Frequency-Reciprocal space on a Plane Waves basis set.
! Calculate Dielectric Properties.

!==================================================================
subroutine check_kramerskronig(n,o,eps)
!==================================================================
! check Kramers Kronig
! \int_0^\infty d\omega' frac{\omega'}{\omega'^2 - \omega^2} 
! Im \epsilon(\omega') = Re \epsilon(\omega)
!------------------------------------------------------------------
      use ggwdef
      implicit none
      integer n
      real o(n)
      complex eps(n)
      complex int(n)
      real e1kk(n)
      integer i, ip
      real omega, omegap, domega
      complex c
      real kk, kkrms, eav
      complex, external :: simpson

!     calculate domega step and verify all  
      domega = (o(n) - o(1)) / (n-1)
      do i = 2, n
!       if domega is not constant return
        if(domega-(o(i)-o(i-1)) > 0.001) return
      enddo
!     if omega(1) is not 0 return
      if(o(1) > 0.1/HARTREE) return
!     if eps(n) is not 0 warn
      if(aimag(eps(n)) > 0.1) then
        print *, 'warning: Im epsilon for omega = ',o(n)*HARTREE,'eV'
        print *, '         is not yet zero, epsilon_2 = ', aimag(eps(n))
        print *, '         Kramers Kronig test could give wrong results'
      endif
      
!     perform kramers kronig      
      do i = 1, n
        omega = o(i)
        c = (0.0,0.0)
        do ip = 1, n
          if(ip == i) cycle
          omegap = o(ip)
          c = c + omegap / (omegap**2 - omega**2) * aimag(eps(ip))
        enddo
        e1kk(i) = 1.0 + 2.0/PI * domega * real(c)
      enddo
      
!     perform kramers kronig with simpson integration    
      do i = 1, n
        omega = o(i)
        do ip = 1, n
          if(ip == i) cycle
          omegap = o(ip)
          int(ip) = omegap / (omegap**2 - omega**2) * aimag(eps(ip))
        enddo
        c = simpson(n,domega,int)
        e1kk(i) = 1.0 + 2.0/PI * real(c)
      enddo

!     werify kramers kronig
      eav = 0.0
      kk = 0.0
      kkrms = 0.0
      do i = 1, n
        kk = kk + abs(real(eps(i)) - e1kk(i))
        kkrms = kkrms +(real(eps(i)) - e1kk(i))*(real(eps(i)) - e1kk(i))
        eav = eav + abs(real(eps(i)))
      enddo
      eav = eav / n
      kk = (kk / n) / eav
      kkrms = (kkrms / n) / (eav*eav)

      kk = abs(real(eps(1)) - e1kk(1)) / real(eps(1))

!     write data
      write(*,'(" the Kramers-Kronig is verified within ",f7.2,"%")') kk*100
      print *


!     write data      
!      open(22,file='out.kk',status='unknown')
!      write(22,'("# Kramers Kronig calculation of epsilon1")')
!      write(22,'("# omega   epsilon1  epsilon1kk")')
!      do i = 1, n
!        write(22,'(f7.3,2e15.7)') o(i)*HARTREE, real(eps(i)), e1kk(i)
!      enddo
!      close(22)
      
      end
      



!     ==================================================================
      subroutine check_fsumrule(n,o,e2,omegaplasma)
!     ==================================================================
!     check f-sum rule
!     \int_0^\infty d\omega \omega Im \epsilon_GG'(q,\omega) =
!       = \frac{1}{2} \pi \omega_p^2  \frac{\rho(G-G')}{\rho(0)} 
!         versor(q+G) \dot versor(q+G')
!     for q = G = G' = 0, it reads:
!     \int_0^\infty d\omega \omega Im \epsilon_00(q=0,\omega) = 
!       = \pi \omega_p^2 / 2
!     calculate only the second one
!     calculate the integral to evaluate an omega_plasma^eff
!     to compare with omega_plasma
!     ------------------------------------------------------------------
      use ggwdef      
      implicit none
      integer n ! number of omega
      real o(n), e2(n) ! omega and imaginary part of epsilon_00
      real omegaplasma
      integer i, ip
      real omegap, domega
      complex int(n)
      real integral
      real omegaplasmaeff, fsumrule
      complex, external :: simpson
            

!     calculate domega step and verify      
      domega = (o(n) - o(1)) / (n-1)
      do i = 2, n
        if(domega-(o(i)-o(i-1)) > 0.001) return
      enddo
!     if omega(1) is not 0 return
      if(o(1) > 0.1/HARTREE) return
!     if eps(n) is not 0 warn
      if(e2(n) > 0.1) then
        print *, 'warning: Im epsilon for omega = ',o(n)*HARTREE,'eV'
        print *, '         is not yet zero, epsilon_2 = ', e2(n)
        print *, '         f-sum rule test could give wrong results'
      endif
      
!     integrate to obtain f-sum rule
      integral = 0.0
      do ip = 1, n
        omegap = o(ip)
        integral = integral + omegap * e2(ip)
      enddo
      integral = domega * integral
      
!     integrate with simpson to obtain f-sum rule   
      do ip = 1, n
        omegap = o(ip)
        int(ip) = omegap * e2(ip) 
      enddo
      integral = real(simpson(n,domega,int))
      omegaplasmaeff = sqrt(integral*2/PI)

      fsumrule = abs((omegaplasmaeff - omegaplasma)) / omegaplasma

!     write data
      write(*,'(" omega_plasma     = ",f6.2," [eV]")') omegaplasma*HARTREE
      write(*,'(" omega_plasma^eff = ",f6.2," [eV]")') omegaplasmaeff*HARTREE
      write(*,'(" the f-sum rule is verified within ",f6.2,"%")') fsumrule*100
      print *

      return
      end
      




      complex function simpson(n,h,f)
      
      implicit none
      integer n
      real h
      complex f(n)
      
      integer i
      complex sumeven, sumodd

!     if n even, neglect last value
      if((n/2)*2 == n) n = n - 1

      sumodd = 0.0
      do i = 2, n-1, 2
        sumodd = sumodd + f(i)
      enddo
      
      sumeven = 0.0
      do i = 3, n-2, 2
        sumeven = sumeven + f(i)
      enddo
      
      simpson = h/3 * (f(1) + sumodd*4 + sumeven*2 + f(n))
      
      return
      end

